! -----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----
!
! Compilation with (ADAPTOR + MPICH):
!  adaptor -hpf -dm -free -o lumm_ccc.exe lumm_ccc.f90
!
! Running with MPICH:
!  mpdboot -n 21 -f ~/mpd.hosts
!  mpdtrace
!  mpiexec -machinefile ~/machi.dat -np 2 lumm_ccc.exe
!  mpdallexit
!
! -----------------------------------------------------------------
program lumm_ccc_test
  implicit none
  real,    parameter :: eps = 0.001
  integer, parameter :: n = 100     ! nro de incognitas
  !
  real    :: maxdiff
  integer :: errata = 0, i, np
  !
  np = number_of_processors ()
  call lumm_ccc (n, np, maxdiff)
  !
  if (maxdiff .gt. eps) then
    errata = errata + 1
    print *, "--> : result is not exact enough for lumm_ccc ", n
  end if
  if (errata .eq. 0) then
    print *, "lumm, np =", np," : all tests passed"
  else
    print *, "lumm, np =", np," : at least one test failed"
  end if
  !
end program
! -----------------------------------------------------------------
subroutine lumm_ccc (n, np, maxdiff)
  implicit none
  !
  integer, intent (in)   :: n
  integer, intent (out)  :: np
  real   , intent (out)  :: maxdiff
  !
  real,    parameter     :: eps = 0.001
  real, dimension (n,n)  :: a, lu, b
  real, dimension (n)    :: col, row
  !
  real                   :: x
  integer                :: h1
  real, dimension (5)    :: time
  !hpf$ distribute a (cyclic,cyclic)
  !
  !adp$ layout a (global,global)
  !adp$ layout b (global,local)
  !adp$ global :: col, row
  !
  !hpf$ align  lu (i,j) with a (i,j)
  !hpf$ align  b  (i,j) with a (i,j)
  !hpf$ align  col  (i) with a (i,*)
  !hpf$ align  row  (j) with a (*,j)
  !
  integer :: np1 = 0, np2 = 0 
  integer :: i, j, k
  !
  call get_shape (a, np1, np2)
  print *, "proc shape = ", np1, " x ", np2
  !
  call init (a, n)
  lu = a
  !
  print *,"LU decomposition "
  call system_clock (count = h1)   ! t1 = second ()
  !
  do k = 1, (n - 1)
     x = lu (k,k)
     if (abs (x) .lt. eps) then
       print *, "no pivoting, error at k = ", k, " x = ", x
       stop
     end if
     !hpf$ independent
     lu  (k+1:n,k)     = lu (k+1:n,k) / lu (k,k)
     col (k+1:n)       = lu (k+1:n,k)
     row (k+1:n)       = lu (k,k+1:n)
     lu  (k+1:n,k+1:n) = lu (k+1:n,k+1:n)      &
              - spread (col (k+1:n), 2, n-k)   &
              * spread (row (k+1:n), 1, n-k)
  end do ! k
  call  walltime (h1, time (1))
  !
  print *, "LU ready, now finds B such  A * B = I "
  call system_clock (count = h1)   ! t1 = second ()
  !
  b = 0.0
  forall (i=1:n) b (i,i) = 1.0
  do  j = 1, n
    col (j+1:n)   = lu (j+1:n,j)
    row (1:n)     =  b (j,1:n)
    b (j+1:n,1:n) =  b (j+1:n,1:n)      &
         - spread (col (j+1:n), 2, n  ) & 
         * spread (row (1:n)  , 1, n-j)
  end do
  call  walltime (h1, time (2) )
  !
  print *, "now finds X with U * X = I "
  call system_clock (count = h1)   ! t1 = second ()
  !
  do j = n, 1, -1
    col (1:n)   = lu (1:n,j)
    b   (j,1:n) =  b (j,1:n) / col (j)
    row (1:n)   =  b (j,1:n)
    b   (1:j-1,1:n) = b (1:j-1,1:n)   & 
          - spread (col (1:j-1), 2, n) &
          * spread (row (1:n), 1, j-1)
  end do
  call  walltime (h1, time (3) )
  !
  print *, "we have the inverse, now do LU = A * B "
  call system_clock (count = h1)   ! t1 = second ()
  lu = 0
  do  k = 1,n
    col = a (1:n,k)
    row = b (k,1:n)
    lu (1:n,1:n) = lu (1:n,1:n)  &
        + spread (col, 2, n)     &
        * spread (row, 1, n)
  end do
  call  walltime (h1, time (4) )
  !
  print *, "matmul correct, now check = I "
  call system_clock (count = h1)   ! t1 = second ()
  a = lu
  forall (i=1:n) a(i,i) = a(i,i) - 1.0
  a = abs (a)
  maxdiff = maxval (a)
  call  walltime (h1, time (5) )
  !
  print * , "maximal eps = ", maxdiff
  !
  print *, "ciclos    = ", h1
  print *, "LU decomposition ; time [s] = ", time (1)
  print *, "finds B          ; time [s] = ", time (2)
  print *, "finds x          ; time [s] = ", time (3)
  print *, "product A B      ; time [s] = ", time (4)
  print *, "check I          ; time [s] = ", time (5)
  !
end subroutine
! -----------------------------------------------------------------
! a will be set to a matrix with the following entries:
! (this matrix works fine to make a lu decomposition)
! 1.0  1.0  0.0  0.0  0.0
! 1.0  0.0  2.0  0.0  0.0
! 1.0  0.0  0.0  3.0  0.0
! 1.0  0.0  0.0  0.0  4.0
! 1.0  0.0  0.0  0.0  0.0
! -----------------------------------------------------------------
subroutine init (a, n)
  implicit none
  integer, intent (in)  :: n
  real, dimension (n,n) :: a
  !hpf$ distribute (cyclic,cyclic) :: a
  integer :: i
  a = 0.0
  forall (i=1:n-1) a(i,i+1) = real(i)
  a (:,1) = 1.0
end subroutine
! -----------------------------------------------------------------
subroutine get_shape (a, np1, np2)
  use hpf_library
  implicit none
  integer :: np1, np2
  real, dimension (:,:) :: a
  !hpf$ distribute (cyclic,cyclic) :: a
  integer :: shape (2)
  call hpf_distribution (a, processors_shape=shape)
  np1 = shape(1)
  np2 = shape(2)
end subroutine
! -----------------------------------------------------------------
subroutine walltime (h1,time)
  implicit none
  real    :: time
  integer :: h1, h2, h3
  call system_clock (h2, h3)
  h1 = h2 - h1
  time = float (h1) / float (h3)
end subroutine
! =================================================================
