! -----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----
!
! compilation with (adaptor + mpich):
!  adaptor -hpf -dm -free -o lap_indirect.exe lap_indirect.f90
!
! running with mpich:
!  mpdboot -n 21 -f ~/mpd.hosts
!  mpdtrace
!  mpiexec -machinefile ~/machi.dat -np 2 lap_indirect.exe
!  mpdallexit
! -----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----
program laplace
  implicit none
  ! solving the laplace pde
  integer, parameter :: n1 = 64, n2 = 64
  real   , parameter :: eps = 0.01
  integer :: i, j, k
  integer :: n, np
  ! irregular data structures
  real   , allocatable, dimension (:,:) :: f
  real   , allocatable, dimension (:)   :: g1, g2
  logical, allocatable, dimension (:)   :: m
  integer, allocatable, dimension (:,:) :: ind
  !hpf$ distribute (block)      :: g1, g2
  !hpf$ align (i)   with g1 (i) :: m
  !hpf$ align (*,i) with g2 (i) :: ind
  integer :: iter, h1, h2, h3
  real    :: time
  logical :: okay
  !
  n = n1 * n2 
  np = number_of_processors ()
  print 20, n1, n2, np
  20 format ('run problem',i4,' x',i4,' on ',i3, ' processors')
  !
  ! allocating memory for regular case
  allocate (f (n1, n2))
  ! allocate memory for irregular case
  allocate (g1(n),g2(n),m(n),ind(4,n))
  ind = -1
  m   = .false.
  do  j = 2, n2-1
     do  i = 2, n1-1
        k = (j-1)*n1 + i
        ind (1,k) = k+n1    !  corresponds (i,j+1)
        ind (2,k) = k-n1    !  corresponds (i,j-1)
        ind (3,k) = k-1     !  corresponds (i-1,j)
        ind (4,k) = k+1     !  corresponds (i+1,j)
        m(k) = .true.
     end do
  end do
  ! initialization of regular grid f
  f = 2.0
  f (:,n2) = 1.
  forall (j=2:n1-1,i=2:n2-1)
     f (i,j) =  0
  end forall
  ! initialization of irregular grid g
  do j = 1, n2
    k = (j-1)*n1
    g1 (k+1:k+n1) = f (1:n1,j)
  end do
  g2 = g1 
  print *, 'initialization is ready'
  ! now iterate on the grid f
  call system_clock (h1, h3)
  call iterate1 (g1, m, ind, n, eps, iter)
  call system_clock (h2)
  h2 = h2 - h1
  time = float(h2)/float(h3)
  print 10, 1, time, iter
  10  format ('rUn ',i1,' needed ',f7.3,' secs for ',i5,' iterations')
  !
  ! now iterate on the grid g2
  call system_clock (h1, h3)
  call iterate2 (g2, m, ind, n, eps, iter)
  call system_clock (h2)
  h2 = h2 - h1
  time = float (h2)/float(h3)
  print 10, 2, time, iter
  !
  ! now compare the results
  okay = all (g1 == g2)
  if  (okay) then
    print *, 'lap_indirect passed'
   else
    print *, 'lap_indirect failed'
  end if
end program
! -----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----
subroutine  iterate1 (g, m, ind, n, eps, niter)
  implicit none
  integer, intent (in)  :: n
  integer, intent (out) :: niter
  real   , intent (in)  :: eps
  real   , dimension (n)   :: g
  logical, dimension (n)   :: m
  integer, dimension (4,n) :: ind
  real   , dimension (n)   :: dg
  real    :: fmax, s
  integer :: i, j
  !hpf$ distribute (block) :: g
  !hpf$ align (i)   with g (i) :: m, dg
  !hpf$ align (*,i) with g (i) :: ind
  !adp$ trace :: ind
  !
  dg = 0.0
  niter = 0
  fmax = 1.0
  ! iteration
  do while (fmax > eps) 
    niter = niter + 1
    !hpf$ independent, new(s)
    do  i = 1, n
      if (m (i) ) then 
        s = sum (g (ind (:,i)))
        dg (i) =  s * 0.25 - g (i)
      end if
    end do
    g = g + dg
    dg = abs(dg)
    fmax = maxval (dg)
  end do
end subroutine
! -----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----
subroutine iterate2 (g, m, ind, n, eps, niter)
  implicit none
  integer, intent (in)  :: n
  integer, intent (out) :: niter
  real   , intent (in)  :: eps
  real   , dimension (n)   :: g
  logical, dimension (n)   :: m
  integer, dimension (4,n) :: ind
  real   , dimension (n)   :: dg
  real    :: fmax, s
  integer :: i, j
  !hpf$ distribute (block) :: g
  !hpf$ align (i)   with g (i) :: m, dg
  !hpf$ align (*,i) with g (i) :: ind
  !adp$ trace   :: ind
  !hpf$ dynamic :: g
  !
  dg = 0.0
  niter = 0
  fmax = 1.0
  ! iteration
  do while (fmax > eps) 
    niter = niter + 1
    !hpf$ independent, new(s)
    do i = 1, n
      if ( m (i) ) then 
         s = sum (g (ind(:,i)))
         dg (i) =  s * 0.25 - g(i)
      end if
    end do
    g = g + dg
    dg = abs (dg)
    fmax = maxval (dg)
  end do
  !
end subroutine
! -----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----
