! -----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----
!
! compilation with (adaptor + mpich):
!  adaptor -hpf -dm -free -o lap_regular.exe lap_regular.f90
!
! running with mpich:
!  mpdboot -n 21 -f ~/mpd.hosts
!  mpdtrace
!  mpiexec -machinefile ~/machi.dat -np 2 lap_regular.exe
!  mpdallexit
! -----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----
program laplace
  implicit none
  ! solving the laplace pde
  integer, parameter :: n1 = 64, n2 = 64
  real   , parameter :: eps = 0.01
  real   , dimension (n1, n2) :: f1, f2
  !hpf$ distribute (block,block) :: f1, f2
  integer :: iter, h1, h2, h3
  integer :: i, j, p
  real    :: time
  !
  p = number_of_processors ()
  print *, 'run stencil oper. ', n1, ' x ', n2, ' on ', p, ' processors'
  !
  ! initialization
  f1 = 2.
  f1 (:,n2) = 1.
  forall (j=2:n2-1,i=2:n1-1)
     f1 (i,j) =  0
  end forall
  f2 = f1
  !
  ! iterate with temporaries 
  call system_clock (h1, h3)
  call iterate1 (f1, n1, n2, eps, iter)
  call system_clock (h2)
  !
  h2 = h2 - h1
  time = float (h2) / float (h3)
  print 10, 'run (temporaries)', time, iter
  10  format (a25,' needed ',f7.3,' secs for ',i5,' iterations')
  !
  ! now iterate with shadow edges
  call system_clock (h1, h3)
  call iterate2 (f2, n1, n2, eps, iter)
  call system_clock (h2)
  h2 = h2 - h1
  time = float (h2) / float(h3)
  print 10, 'run (shadow)     ', time, iter
  !
  ! compare the results
  if (all (f1 == f2)) then
    print *, 'lap_regular passed'
   else
    print *, 'lap_regular failed'
  end if
  !
end program
! -----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----
subroutine iterate1 (f, n1, n2, eps, niter)
  implicit none
  integer, intent (in)  :: n1, n2
  real   , intent (in)  :: eps
  integer, intent (out) :: niter
  real   , dimension (n1,n2) :: f, df
  !hpf$ distribute f (block,block)
  !hpf$ align df (:,:) with f (:,:)
  !hpf$ shadow f (0:0,0:0)          ! allows no shadow
  real    :: fmax
  integer :: i, j
  df = 0.0
  niter = 0
  fmax = 1.0
  ! iteration
  do while (fmax > eps)
     niter = niter + 1
     forall (j=2:n2-1,i=2:n1-1)
       df(i,j) = ( f (i,j+1) + f (i,j-1) &
               +   f (i-1,j) + f (i+1,j) ) * 0.25 - f (i,j)
     end forall
     forall (j=2:n2-1,i=2:n1-1)
        f (i,j) = f (i,j) + df (i,j)
     end forall
     df = abs (df)
     fmax = maxval (df)
  end do
  !
end subroutine
! -----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----
!  improved version 
! -----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----
subroutine iterate2 (f, n1, n2, eps, niter)
  implicit none
  integer, intent (in)  :: n1, n2
  real   , intent (in)  :: eps
  integer, intent (out) :: niter
  real   , dimension (n1,n2) :: f, df
  !hpf$ distribute f (block,block)
  !hpf$ align df (:,:) with f (:,:)
  real    :: fmax
  integer :: i, j
  !
  df    = 0.0
  niter = 0
  fmax  = 1.0
  !
  ! iteration
  do while (fmax > eps)
     niter = niter + 1
     forall (j=2:n2-1,i=2:n1-1)
       df (i,j) = ( f (i,j+1) + f (i,j-1)  &
                +   f (i-1,j) + f (i+1,j) ) * 0.25 - f (i,j)
     end forall
     forall (j=2:n2-1,i=2:n1-1)
        f (i,j) = f (i,j) + df (i,j)
     end forall
     df = abs (df)
     fmax = maxval (df)
  end do
end subroutine
! -----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----
