! -----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----
! Compilation with INTEL:
!  ifort -w -o gauss.exe2 gauss_dyna.f90
!
! Compilation with GNU:
!  g95 -w -o gauss.exe2 gauss_dyna.f90
!
! Compilation with (ADAPTOR + MPICH):
!  adaptor -hpf -dm -free -o gauss.exe2 gauss_dyna.f90
!
! Running with MPICH:
!  mpdboot -n 21 -f ~/mpd.hosts
!  mpdtrace
!  mpiexec -machinefile ~/machi.dat -np 2 gauss.exe2
!  mpiexec -machinefile ~/machi.dat -1 -np 4 gauss.exe2
!  mpdallexit
!
! The same running but with MPI (old):
!  mpirun -machinefile ~/machi.dat -np 2 gauss.exe2
!  mpirun -machinefile ~/machi.dat -nolocal -np 4 gauss.exe2
!
! Some running details:
!  si n = 8 000 entonces m = n^2, RAM = m * 8 /1e6 =  512 [Mbytes]
!  pero como aqui hay una copia de A se tiene ram = 1,024 [Gbytes]
!
! output example:
!  gauss 
!  number of unknowns ; n = ? 
!  max diff =  5.83895598538220E-10
!  time   =  576.0575  seconds
!  mflops =  374.9626  on  8  processors
!  mflops =  46.87032  for one processor
! -----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----
program gauss_dyna
  implicit none
  integer :: n
  integer :: i, j, k, j1
  logical, parameter :: verifica = .false. 
  double precision, allocatable :: a  (:,:), b  (:), x (:)
  double precision, allocatable :: a0 (:,:), b0 (:), r (:)
   !hpf$ distribute  (*,block) :: a   , a0
   !hpf$ distribute    (block) :: b, x, b0, r
  !!hpf$ distribute (*,cyclic) :: a   , a0
  !!hpf$ distribute   (cyclic) :: b, x, b0, r
  double precision, allocatable :: ak (:)
  double precision :: q, diff
  double precision :: bk
  character (11) :: arch1
  character (4)  :: exte1
  character (5)  :: raiz
  real    :: time, mflops
  integer :: tickstart, tickstop, tickrate, kciclos
  integer :: l1, np
  !
  ! define archivos
  raiz = "gauss"
  l1 = len_trim (raiz)
  exte1 = ".tim"
  arch1 = raiz (1:l1) // exte1
  l1 = len_trim (arch1)
  !
  !carteles
  write (*,*)
  write (*,*)" gauss solution of a SEL, with dynamic memory"
  write (*,*)
  write (*,*)" number of unknowns ; n = ? "
  read (*,*) n
  !
  !aloca memoria dinamica
  allocate ( a  (n,n), b  (n), x (n) )
  allocate ( a0 (n,n), b0 (n), r (n) )
  allocate ( ak (n) )
  !
  !genera un SEAL regular
  !adp$ profile init begin
  b = [ (i, i=1,n) ]
  x = 1.0d0
  !hpf$ independent
  do  i = 1, n 
  !hpf$ independent
    do  j = 1, n 
      if (i == j) then
        a (j,i) =  3.0d0
      elseif (i == j-4 .or. j+4 == i)then
        a (j,i) = 18.0d0
      else
        a (j,i) = 0.0d0
      end if
    end do ! j
  end do ! i
  !adp$ end profile
  !
  if  (verifica) then
    b0 = b
    a0 = a
  end if
  !
  !empieza la zona cronometrada
  call system_clock (tickstart)
  !
  !triangulariza
  !adp$ profile gauss
  !hpf$ parallel
  do  k = 1, (n - 1)
    !hpf$ new (ak,q) begin
    ak (k:n) = a (k:n,k)
    q = 1.0d0 / ak (k)
    !hpf$ independent, new (i), on home ( a (:,j) )
    do  j  = (k + 1), n
      b (j) = b (j) - ak (j) * b (k) * q
      do  i = (k + 1), n
        a (i,j) = a (i,j) - ak (i) * a (k,j) * q
      end do ! i
    end do ! j
    !hpf$ end new
  end do ! k
  ! 
  !retrosustitucion
  !adp$ profile backsub begin
  !hpf$ independent
  do  i = 1, n
    x (i) = b (i)
  end do
  x (n) = x (n) / a (n,n)
  do  i = (n - 1), 1, -1
    !hpf$ independent
    do  j = 1, i
      x (j) = x (j) - x (i+1) * a (j,i+1)
    end do
    x (i) = x (i) / a (i,i)
  end do
  !adp$ end profile
  !
  !termina la zona cronometrada
  call system_clock (tickstop, tickrate)
  !
  !verifica la solucion obtenida
  if  (verifica) then
    r = matmul (a0,x)             ! asi es paralelizada automaticamente
    diff = maxval (abs (b0 - r))
    write (*,*)
    write (*,*)" max diff = ", diff 
  end if
  !
  !estadistica
  kciclos = tickstop - tickstart
  time = float (kciclos) / float(tickrate)
  mflops = real (n) ** 3 / (1000000.0 * time)
  np = number_of_processors ()
  !
  ! imprime resultados
  write (*,*)
  write (*,*)" time     = ", time, " seconds"
  write (*,*)" mflops   = ", mflops, " on ", np, " processors"
  write (*,*)" mflops   = ", mflops/np, " for one processor (estimated!)"
  write (*,*)
  !
  ! imprime en archivo historico
  open (12, file = arch1 (1:l1), status = 'unknown')
  do 
    read (12,*,end=50) i
  end do
  50 write (12,400)  np, n, 3, time, time, time
  close (12, status = 'keep')
  400 format (1x, i2, 1x, i8, 1x, i2, 3 (1x, e24.12) )
  !
  ! daloca memoria dinamica
  deallocate (a, b, x)
! deallocate (b0, a0, r)
  deallocate (ak)
  !
end program
! -----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----
