! -----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----
!   Gauss con (A, B), donde
!   A ; matriz de coeficientes
!   B ; matriz de cargas
! -----------------------------------------------------------------
module gaussol_m
  use  constan_m
  use  globals_m
  use  herrami_m
  implicit none
  private                           ! por omision todo es privado
  interface gauss                   ! solucion LU paralelo
    module procedure gauss_d, gauss_z
  end interface
  public gauss
  character (11), parameter :: c = "gaussol_m> "
  contains
  !
  !----------------------------------------------------------------
  !   Triangularizacion Gaussiana con pivotaje: variable real
  subroutine gauss_d (a, b)
    real (idp), dimension (:,:), intent (inout) :: a, b
    real (idp), dimension   (:), allocatable    :: u, v
    real (idp) :: s
    !hpf$  inherit, distribute (*,cyclic) :: a
    !hpf$  align  b (*,:) with  a (*,:)
    integer (iin), dimension (1) :: ipivot
    integer (iin) :: n, m
    integer (iin) :: i, j, k, p
    !
    n = siza (size(b,2), size(a,1), size(a,2), "gauss_d" )
    m = size (b,1)
    !
    ! aloca
    ier = 0
    allocate ( u (n), stat = ier (1) )
    allocate ( v (m), stat = ier (2) )
    if ( any (ier .ne. 0) ) call errata (ier, "gauss: aloca")
    !
    ! inicia cronometro;  
    time = cero
    call system_clock (count = h1)   ! t1 = second ()
    !
    ! lazo Gauss con pivotaje
    gauss: do  k = 1, (n - 1)
      ! selecciona pivote
      ipivot = maxloc ( abs (a(k:n,k)) ) + (k - 1)
      p = ipivot (1)
      ! permuta filas: a (p,:) <--> a (k,:)
      !                b (p)   <--> b (k)
      if  ( p .ne. k ) then
        !hpf$ independent
        forall (j = k : n)  
          u (j)   = a (p,j)
          a (p,j) = a (k,j)
          a (k,j) = u (j)
        end forall
        forall (j = 1 : m)  
          v (j)   = b (j,p)
          b (j,p) = b (j,k)
          b (j,k) = v (j)
        end forall
      end if
      ! triangulariza matriz
      !hpf$ independent
      forall (i = k + 1 : n)
        a (i,k) = a (i,k) / a (k,k)
        !hpf$  independent
        forall (j = k + 1 : n)
          a (i,j) = a (i,j) - a (i,k) * a (k,j)
        end forall
        forall (p = 1 : m)
          b (p,i) = b (p,i) - a (i,k) * b (p,k)
        end forall
      end forall
    end do gauss
    !
    ! delta de tiempo por triangularizar
    call system_clock (count = h2)     ! t2 = second ()
    time (1) = h2 - h1                 ! time (1) = t2 - t1
    !
    ! retro-substitucion
    pp: do  p = 1, m
      ii: do  i = n, 1, -1
        s = cero
        !hpf$ independent, reduction (s)
        jj : do  j = (i + 1), n
          s = s + a (i,j) * b (p,j)
        end do jj
        b (p,i) = ( b (p,i) - s ) / a (i,i)
      end do ii
    end do pp
    !
    ! delta de tiempo por retrosustitucion
    call system_clock (count = h3)     ! t2 = second ()
    time (2) = h3 - h2                 ! time (2) = t3 - t2
    !
    ! estima MFlops
    call mflops (n, m, time)
    !
    ! dloca
    ier = 0
    deallocate ( u, stat = ier (1) )
    deallocate ( v, stat = ier (2) )
    if ( any(ier.ne.0) ) call errata (ier, "gauss: dloca")
  end subroutine
  !
  !----------------------------------------------------------------
  !   Triangularizacion Gaussiana con pivotaje: variable compleja
  subroutine gauss_z (a, b)
    complex (idp), dimension (:,:), intent (inout) :: a, b
    complex (idp), dimension   (:), allocatable    :: u, v
    complex (idp) :: s
    !hpf$ inherit, distribute (*,cyclic) :: a
    !hpf$ align  b (*,:) with  a (*,:)
    real    (idp), dimension (6) :: time
    integer (iin), dimension (2) :: ier
    integer (iin)                :: h1, h2, h3
    integer (iin), dimension (1) :: ipivot
    integer (iin) :: n, m
    integer (iin) :: i, j, k, p
    !
    n = siza (size(b,2), size(a,1), size(a,2), "gauss_d" )
    m = size (b,1)
    !
    ! aloca
    ier = 0
    allocate ( u (n), stat = ier (1) )
    allocate ( v (m), stat = ier (2) )
    if ( any (ier .ne. 0) ) call errata (ier, "gauss: aloca")
    !
    ! inicia cronometro;  
    time = cero
    call system_clock (count = h1)   ! t1 = second ()
    !
    ! lazo Gauss con pivotaje
    gauss: do  k = 1, (n - 1)
      ! selecciona pivote
      ipivot = maxloc ( abs (a(k:n,k)) ) + (k - 1)
      p = ipivot (1)
      ! permuta filas: a (p,:) <--> a (k,:)
      !                b (p)   <--> b (k)
      if  ( p .ne. k ) then
        !hpf$ independent
        forall (j = k : n)  
          u (j)   = a (p,j)
          a (p,j) = a (k,j)
          a (k,j) = u (j)
        end forall
        forall (j = 1 : m)  
          v (j)   = b (j,p)
          b (j,p) = b (j,k)
          b (j,k) = v (j)
        end forall
      end if
      ! triangulariza matriz
      !hpf$ independent
      forall (i = k + 1 : n)
        a (i,k) = a (i,k) / a (k,k)
        !hpf$  independent
        forall (j = k + 1 : n)
          a (i,j) = a (i,j) - a (i,k) * a (k,j)
        end forall
        forall (p = 1 : m)
          b (p,i) = b (p,i) - a (i,k) * b (p,k)
        end forall
      end forall
    end do gauss
    call system_clock (count = h2)     ! t2 = second ()
    time (1) = h2 - h1                 ! time (1) = t2 - t1
    !
    ! retro-substitucion
    pp: do  p = 1, m
      ii: do  i = n, 1, -1
        s = cero
        !hpf$ independent, reduction (s)
        jj : do  j = (i + 1), n
          s = s + a (i,j) * b (p,j)
        end do jj
        b (p,i) = ( b (p,i) - s ) / a (i,i)
      end do ii
    end do pp
    !
    ! estima MFlops
    call system_clock (count = h3)     ! t2 = second ()
    time (2) = h3 - h2                 ! time (2) = t3 - t2
    !
    ! dloca
    ier = 0
    deallocate ( u, stat = ier (1) )
    deallocate ( v, stat = ier (2) )
    if ( any(ier.ne.0) ) call errata (ier, "gauss: dloca")
  end subroutine
! +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!   estadistica adaptada del Linpack Benchmark
! -----------------------------------------------------------------
  subroutine  mflops (n, m, time)
    real    (idp), parameter :: cray = 0.056d0
    real    (idp), parameter :: dos  = 2.0d0
    real    (idp), parameter :: tres = 3.0d0
    real    (idp), parameter :: mega = 1.0d6
    integer (iin),                intent (in)   :: n
    integer (iin),                intent (in)   :: m
    real    (idp), dimension (:), intent (inout):: time
    real    (idp) :: total, ops
    integer (iin) :: p, ciclos
    !
    call system_clock (count_rate = h4)  ! para pasar a segundos
    time (3) = time (1) + time (2)       ! ciclos total
    ciclos = floor (time (3))
    !
    time  = time / h4                    ! pasa de ciclos a seg
    total = time (3)                     ! total en seg
    !
    ops = (dos / tres) * dfloat (n)**3 + dos * dfloat (n)**2
    p = size (time)
    if (p < 7) call errata ("error (mflops): p < 7")
    time (3) = total                 ! tiempo [seg]
    time (4) = ops / (mega * total)  ! Mflops
    time (5) = dos / time (4)        ! unit
    time (6) = total / cray          ! ratio
    !
    ! solo en HPF
    p = number_of_processors ()
    !
    write (*,*)
    write (*,*)"nro de procesadores ;      p  = ", p
    write (*,*)"tamanio de la matriz;      n  = ", n
    write (*,*)"nro de cargas       ;      m  = ", m
    write (*,*)"nro de ciclos       ;  ciclos = ", ciclos
    write (*,*)"tiempo total        ; seconds = ", real (time(3))
    write (*,*)"                       MFlops = ", real (time(4))
    write (*,*)
    write (*,*)"numero de operaciones ; o  = ", real (ops)
    write (*,*)"tpo de factoriz. [seg]; t1 = ", real (time(1))
    write (*,*)"tpo de restrosu. [seg]; t2 = ", real (time(2))
    write (*,*)"unit                       = ", real (time(5))
    write (*,*)"ratio                      = ", real (time(6))
    write (*,*)
    write (*,*)" p  A_size  B_loads    ciclos     seconds      Mflops"
    !
    ! resumen a disco
    open (1, file = 'gauss.tpo', &
             status = 'unknown', &
             position = 'append')
    write (*,110)  p, n, m, ciclos, total, time (4)
    write (1,110)  p, n, m, ciclos, total, time (4)
    close (1, status = 'keep')
    write (*,*)
    !
    ! formatos
    100 format (a)
    110 format (1x, i2, i8, 1x, i8, 1x, i9, 1x, e11.4, 1x, e11.4)
  end subroutine
end module
! -----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----
