  function [ x r K E ] = cgs (A, b)
% function [ x r K E ] = cgs (A, b)
%
% resolucion iterativa del sistema lineal A x = b por
% Gradient Conjugate Squared (CGS)
%
% entran:
%   A     ; matriz del sistema lineal (supuesta no singular)
%   b     ; vector fuente del sistema lineal
%
% salen:
%   x     ; vector solucion iterado
%   r     ; vector residuo  iterado
%   K     ; historia de las iteraciones realizadas
%   E     ; historia de las normas L2 de los vectores residuos

%  previo en las dimensiones de los arreglos
[ n  m1 ] = size (A) ;
m2 = length (b) ;
if  (m1 == m2 & m1 == n),
  m = m1 ;
else
  error ('(cgs): m1 .ne. m2 .or. m1 .ne. n')
end % if

% define parametros en la solucion iterativa
o    = zeros (n,1)   ; % vector nulo
x0   = o             ; % solucion k-1 inicial
tol  = 1.0e-12       ; % tolerancia en la norma del vector residuo
emax = 1.0e+12       ; % residuo maximo admisible
kmax = 8192          ; % maximo numero de iteraciones previsto
k    = 1             ; % inicia contador de iteraciones
E    = 0             ; % inicia historia norma l2 
K    = 0             ; % inicia historia iteraciones
x    = x0            ; % solucion k-1 inicial
Q    = eye (n,n)     ; % precondicionador identidad

% previo
alpha = sum (b .* b) ;
zeta  = abs (alpha)
if  ( zeta < tol ), 
  disp (' aviso: vector fuente nulo')
  return
end % if

p3    = o ;
r3    = o ;
beta  = 0 ;

t1    = A * x ;
t2    = b - t1 ;
r2    = Q * t2 ;
r1    = r2 ;

delta = sum (r2 .* r2)
rho   = sqrt (delta) / alpha
gamma = delta 

% lazo iterativo
while ( k <= kmax & rho > tol ),
  k
  p2    = r2 + beta * r3 ;
  p1    = r3 + beta * p3 ;
  p3    = p2 + beta * p1 ;
  t1    = A * p3 ; 
  t2    = Q * t1 ;

  omega = sum ( r1 .* t2 ) ;
  zeta  = abs (omega) ;
  if  ( zeta < tol ), 
    disp (' aviso: omega nulo')
    break
  else
    tau = gamma / omega ;
  end % if

  r3 = p2 - tau * t2 ;
  p1 = p2 +       r3 ;
  x  = x  + tau * p1 ;
  t1 = A * p1 ;
  t2 = Q * t1 ;
  r2 = r2 - tau * t2 ;

  delta = sum ( r1 .* r2 ) ;
  epsil = sum ( r2 .* r2 ) ;
  rho   = sqrt ( epsil / alpha )
  beta  = delta / gamma ;
  gamma = delta ;

  K  = [ K ;   k ] ; % apendiza 
  E  = [ E ; rho ] ; % apendiza
  k  = k + 1 ;
  if  ( rho > emax ), break, end % if 
end % p
K (1) = [];
E (1) = [];
r     = r1;

sprintf (' tolerancia prefijada              ; tol = %g', tol)
sprintf (' nro de iteraciones efectuado      ;   k = %g', k  )
sprintf (' norma L2 del ultimo vector residuo; rho = %g', rho)
return
