!! calculating PI using one-sided communication with loos synchronization and
!! .. MPI_Alloc_mem
!! .. the usage of MPI_Alloc_mem to allocate the memory window may improve
!! .. performance but needs either Cray pointers or the Fortran 2003 C compatibility
!! .. features. Standard Fortran90/95 is not sufficient.
!! .. MPI_Comm_group, MPI_Group_excl, MPI_Group_difference, MPI_Group_rank, MPI_Group_free
!! .. MPI_Type_extent, MPI_Alloc_mem, MPI_Free_mem 
!! .. MPI_Win_post, MPI_Win_start, MPI_Win_wait, MPI_Win_complete
!! .. MPI_Get_address

program onesidedsophisticated

  use, intrinsic :: iso_c_binding
  implicit none

  include 'mpif.h'

  integer   ::         myrank, nprocs, n, i, islave, master=0, count, win
  integer   ::         status(MPI_STATUS_SIZE)
  integer   ::         ierr, resultlen, tag, nbytes, address
  integer   ::         group_rank, tmp_group, target_group, origin_group
  integer   ::         shape(1)
  character (LEN=MPI_MAX_PROCESSOR_NAME) :: hostname
  real*8    ::         t0, t1, t2, fh
  real*8, dimension(:), pointer :: fi

  ! MPI 2 prescribes to use non standard Cray pointers in Fortran in order to 
  ! allocate the window with MPI_Alloc_mem used for one-sided communication.
  ! The usage of MPI_Alloc_mem may improve the performance.
  ! But arrays pointed to by Cray pointers may not be dynamically allocated.
  ! A bypass is to declare an array which is sufficiently large and use a Cray
  ! pointer. Then the allocation of this "fixed" array is deferred 
  ! until MPI_Alloc_mem is used.
  ! real*8, dimension(200000000) :: fi
  ! pointer ( p, fi )
  !
  ! With Fortran 2003 the C compatibility features allow a proper allocation
  ! of an dynamic array inside a C function using the function
  ! c_f_pointer to convert a C pointer to a Fortran pointer
  type(c_ptr) p
  real*8, parameter :: PI25DP = 3.141592653589793238462643d0
  ! integer chunk, ierr, ilen, rest, is, ie
  real*8 :: mypi  
  double precision f, a, h, sum, x, pi

  f(a) = 4.d0 / (1.d0 + a*a) ! function to integrate
  
  ! Initialize MPI
  Call MPI_Init(Ierr)

  ! Get my rank id and number of MPI processes
  call MPI_Comm_Rank(MPI_COMM_WORLD, myrank, ierr)
  call MPI_Comm_Size(MPI_COMM_WORLD, nprocs, ierr)
  
  ! Get hostname
  call MPI_Get_processor_name(hostname, resultlen, ierr)
  write (*,*) "Rank ", myrank, " out of ", nprocs,  &
    " processes running on ", hostname(1:index(hostname,".")-1)

  
  ! Synchonization point (not really necessary here)
  t0 = MPI_Wtime()
  call MPI_Barrier (MPI_COMM_WORLD, ierr)
  t1 = MPI_Wtime()

  !-----------------------------------------------------------
  ! Do work here ...

  ! two processor groups - for origin and target of the remote memory accesses - 
  ! are built
  call MPI_Comm_group ( MPI_COMM_WORLD, tmp_group, ierr )
  call MPI_Group_excl ( tmp_group, 1, master, origin_group, ierr ) ! all slaves
  call MPI_Group_difference ( tmp_group, origin_group, target_group, ierr )  ! master only
  call MPI_Group_rank ( origin_group, group_rank, ierr )
  write(*,*) "my rank in the origin_group is ", group_rank
  call MPI_Group_rank ( target_group, group_rank, ierr )
  write(*,*) "my rank in the target_group is ", group_rank
  call MPI_Barrier ( MPI_COMM_WORLD, ierr )

  do
  
    ! The master reads an integer number form standard input and sends it to all slaves
    if ( myrank .eq. master ) then ! Master part
      write(*,*) 'Number of intervals ? (0 quits): '
      read (*,*) n
    end if
    call MPI_Bcast ( n, 1, MPI_INTEGER, master, MPI_COMM_WORLD, ierr )
    if ( n .le. 0 .or. n > 10000000 ) exit    
    
    if ( myrank .eq. master ) then ! Master part
      call MPI_Type_extent ( MPI_DOUBLE_PRECISION, nbytes, ierr )

      ! MPI_Alloc_mem delivers a C-pointer which corresponds to a Cray-pointer 
      ! in Fortran 2003 it can be transformed into a Fortran pointer
      call MPI_Alloc_mem ( n*nbytes, MPI_INFO_NULL, p, ierr )
      shape(1) = n
      call c_f_pointer ( p, fi, shape )

      print *, 'size of fi:', size(fi)

      call MPI_Win_create ( fi, nbytes*n, nbytes, MPI_INFO_NULL, MPI_COMM_WORLD, win, ierr )
    else
      call MPI_Win_create ( fi, 0, 8, 0, MPI_COMM_WORLD, win, ierr )
    end if
    call MPI_Bcast ( n, 1, MPI_INTEGER, master, MPI_COMM_WORLD, ierr )
          

    if ( myrank .eq. master ) then ! Master part
      call MPI_Win_post ( origin_group, 0, win, ierr )
    else
      call MPI_Win_start ( target_group, 0, win, ierr ) 
    end if
    
     h = 1.0d0/dble(n)   ! even step size h as a function of partitions
     do i = myrank+1, n, nprocs  ! simple approach of work sharing
	x = h * (dble(i) - 0.5d0)
	fh = f(x)
        if ( myrank .eq. master ) then ! Master part
	  fi(i) = fh
        else
	  call MPI_Put ( fh, 1, MPI_DOUBLE_PRECISION, master, i-1, 1, MPI_DOUBLE_PRECISION, win, ierr)
        end if
     end do

     if ( myrank .eq. master ) then ! Master part
       call MPI_Win_wait ( win, ierr )
     else
       call MPI_Win_complete ( win, ierr ) 
     end if
     call MPI_Get_address ( fi, address, ierr ) ! dummy call to suppress certain compiler optimizations
 
     if ( myrank .eq. 0 ) then
        sum = 0.0d0
	do i = 1, n
	   sum = sum + fi(i)
	end do
	pi = h * sum
        write(*,'(a,F18.16,a,F18.16)') 'pi is approximately: ',pi, &
             ' Error is: ', abs(pi - PI25DP)
	!deallocate ( fi )
        call MPI_Win_free ( win, ierr )
	call MPI_Free_mem ( fi, ierr )
     else
        call MPI_Win_free ( win, ierr )
     end if
    
  end do
  
  call MPI_Group_free ( tmp_group, ierr )
  call MPI_Group_free ( origin_group, ierr )
  call MPI_Group_free ( target_group, ierr )
     
  !-----------------------------------------------------------

  ! Synchronization point (not really necessary here)
  call MPI_Barrier (MPI_COMM_WORLD, ierr)
  t2 = MPI_Wtime()
  write (*,*) "process ", myrank, ": ", t2-t1, " seconds"

  ! Shut Down MPI
  call MPI_Finalize(ierr )

end program onesidedsophisticated
